#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from tempfile import NamedTemporaryFile

import networkx as nx
import numpy as np

from miplearn.h5 import H5File
from miplearn.problems.stab import (
    MaxWeightStableSetData,
    build_stab_model_pyomo,
    build_stab_model_gurobipy,
)


def test_stab() -> None:
    data = MaxWeightStableSetData(
        graph=nx.cycle_graph(5),
        weights=np.array([1.0, 1.0, 1.0, 1.0, 1.0]),
    )
    for model in [
        build_stab_model_pyomo(data),
        build_stab_model_gurobipy(data),
    ]:
        with NamedTemporaryFile() as tempfile:
            with H5File(tempfile.name) as h5:
                model.optimize()
                model.extract_after_mip(h5)
                assert h5.get_scalar("mip_obj_value") == -2.0



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from tempfile import NamedTemporaryFile

import numpy as np
from scipy.stats import randint, uniform

from miplearn.h5 import H5File
from miplearn.problems.setcover import (
    SetCoverData,
    build_setcover_model_gurobipy,
    SetCoverGenerator,
    build_setcover_model_pyomo,
)


def test_set_cover_generator() -> None:
    np.random.seed(42)
    gen = SetCoverGenerator(
        n_elements=randint(low=3, high=4),
        n_sets=randint(low=5, high=6),
        costs=uniform(loc=0.0, scale=100.0),
        costs_jitter=uniform(loc=0.95, scale=0.10),
        density=uniform(loc=0.5, scale=0),
        K=uniform(loc=25, scale=0),
        fix_sets=False,
    )
    data = gen.generate(2)

    assert data[0].costs.round(1).tolist() == [136.8, 86.2, 25.7, 27.3, 102.5]
    assert data[0].incidence_matrix.tolist() == [
        [1, 0, 1, 0, 1],
        [1, 1, 0, 0, 0],
        [1, 0, 0, 1, 1],
    ]
    assert data[1].costs.round(1).tolist() == [63.5, 76.6, 48.1, 74.1, 93.3]
    assert data[1].incidence_matrix.tolist() == [
        [1, 1, 0, 1, 1],
        [0, 1, 0, 1, 0],
        [0, 1, 1, 0, 0],
    ]


def test_set_cover_generator_with_fixed_sets() -> None:
    np.random.seed(42)
    gen = SetCoverGenerator(
        n_elements=randint(low=3, high=4),
        n_sets=randint(low=5, high=6),
        costs=uniform(loc=0.0, scale=100.0),
        costs_jitter=uniform(loc=0.95, scale=0.10),
        density=uniform(loc=0.5, scale=0.00),
        fix_sets=True,
    )
    data = gen.generate(3)

    assert data[0].costs.tolist() == [136.75, 86.17, 25.71, 27.31, 102.48]
    assert data[1].costs.tolist() == [135.38, 82.26, 26.92, 26.58, 98.28]
    assert data[2].costs.tolist() == [138.37, 85.15, 26.95, 27.22, 106.17]

    print(data[0].incidence_matrix)

    for i in range(3):
        assert data[i].incidence_matrix.tolist() == [
            [1, 0, 1, 0, 1],
            [1, 1, 0, 0, 0],
            [1, 0, 0, 1, 1],
        ]


def test_set_cover() -> None:
    data = SetCoverData(
        costs=np.array([5, 10, 12, 6, 8]),
        incidence_matrix=np.array(
            [
                [1, 0, 0, 1, 0],
                [1, 1, 0, 0, 0],
                [0, 0, 1, 1, 1],
            ],
        ),
    )
    for model in [
        build_setcover_model_pyomo(data),
        build_setcover_model_gurobipy(data),
    ]:
        with NamedTemporaryFile() as tempfile:
            with H5File(tempfile.name) as h5:
                model.optimize()
                model.extract_after_mip(h5)
                assert h5.get_scalar("mip_obj_value") == 11.0



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import Tuple

import numpy as np

from miplearn.h5 import H5File


def _extract_bin_var_names_values(
    h5: H5File,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    bin_var_names, bin_var_indices = _extract_bin_var_names(h5)
    var_values = h5.get_array("mip_var_values")
    assert var_values is not None
    bin_var_values = var_values[bin_var_indices].astype(int)
    return bin_var_names, bin_var_values, bin_var_indices


def _extract_bin_var_names(h5: H5File) -> Tuple[np.ndarray, np.ndarray]:
    var_types = h5.get_array("static_var_types")
    var_names = h5.get_array("static_var_names")
    assert var_types is not None
    assert var_names is not None
    bin_var_indices = np.where(var_types == b"B")[0]
    bin_var_names = var_names[bin_var_indices]
    assert len(bin_var_names.shape) == 1
    return bin_var_names, bin_var_indices



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from setuptools import setup, find_namespace_packages

setup(
    name="miplearn",
    version="0.3.0.dev0",
    author="Alinson S. Xavier",
    author_email="axavier@anl.gov",
    description="Extensible Framework for Learning-Enhanced Mixed-Integer Optimization",
    url="https://github.com/ANL-CEEESA/MIPLearn/",
    packages=find_namespace_packages(),
    python_requires=">=3.7",
)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import numpy as np
from scipy.stats import uniform, randint

from miplearn.problems.multiknapsack import (
    MultiKnapsackGenerator,
    MultiKnapsackData,
    build_multiknapsack_model,
)


def test_knapsack_generator() -> None:
    np.random.seed(42)
    gen = MultiKnapsackGenerator(
        n=randint(low=5, high=6),
        m=randint(low=3, high=4),
        w=randint(low=0, high=1000),
        K=randint(low=500, high=501),
        u=uniform(loc=0.0, scale=1.0),
        alpha=uniform(loc=0.25, scale=0.0),
        fix_w=True,
        w_jitter=uniform(loc=0.9, scale=0.2),
        p_jitter=uniform(loc=0.9, scale=0.2),
        round=True,
    )
    data = gen.generate(2)
    assert data[0].prices.tolist() == [433.0, 477.0, 802.0, 494.0, 458.0]
    assert data[0].capacities.tolist() == [458.0, 357.0, 392.0]
    assert data[0].weights.tolist() == [
        [111.0, 392.0, 945.0, 276.0, 108.0],
        [64.0, 633.0, 20.0, 602.0, 110.0],
        [510.0, 203.0, 303.0, 469.0, 85.0],
    ]

    assert data[1].prices.tolist() == [344.0, 527.0, 658.0, 519.0, 460.0]
    assert data[1].capacities.tolist() == [449.0, 377.0, 380.0]
    assert data[1].weights.tolist() == [
        [92.0, 473.0, 871.0, 264.0, 96.0],
        [67.0, 664.0, 21.0, 628.0, 129.0],
        [436.0, 209.0, 309.0, 481.0, 86.0],
    ]


def test_knapsack_model() -> None:
    data = MultiKnapsackData(
        prices=np.array([344.0, 527.0, 658.0, 519.0, 460.0]),
        capacities=np.array([449.0, 377.0, 380.0]),
        weights=np.array(
            [
                [92.0, 473.0, 871.0, 264.0, 96.0],
                [67.0, 664.0, 21.0, 628.0, 129.0],
                [436.0, 209.0, 309.0, 481.0, 86.0],
            ]
        ),
    )
    model = build_multiknapsack_model(data)
    model.optimize()
    assert model.inner.objVal == -460.0



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from math import log
from typing import List, Dict, Any
import numpy as np

import gurobipy as gp

from ..h5 import H5File


class ExpertBranchPriorityComponent:
    def __init__(self) -> None:
        pass

    def fit(self, train_h5: List[str]) -> None:
        pass

    def before_mip(self, test_h5: str, model: gp.Model, _: Dict[str, Any]) -> None:
        with H5File(test_h5, "r") as h5:
            var_names = h5.get_array("static_var_names")
            var_priority = h5.get_array("bb_var_priority")
            assert var_priority is not None
            assert var_names is not None

            for (var_idx, var_name) in enumerate(var_names):
                if np.isfinite(var_priority[var_idx]):
                    var = model.getVarByName(var_name.decode())
                    var.branchPriority = int(log(1 + var_priority[var_idx]))



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from abc import ABC, abstractmethod
from typing import Optional, Dict

import numpy as np

from miplearn.h5 import H5File


class AbstractModel(ABC):
    _supports_basis_status = False
    _supports_sensitivity_analysis = False
    _supports_node_count = False
    _supports_solution_pool = False

    @abstractmethod
    def add_constrs(
        self,
        var_names: np.ndarray,
        constrs_lhs: np.ndarray,
        constrs_sense: np.ndarray,
        constrs_rhs: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        pass

    @abstractmethod
    def extract_after_load(self, h5: H5File) -> None:
        pass

    @abstractmethod
    def extract_after_lp(self, h5: H5File) -> None:
        pass

    @abstractmethod
    def extract_after_mip(self, h5: H5File) -> None:
        pass

    @abstractmethod
    def fix_variables(
        self,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        pass

    @abstractmethod
    def optimize(self) -> None:
        pass

    @abstractmethod
    def relax(self) -> "AbstractModel":
        pass

    @abstractmethod
    def set_warm_starts(
        self,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        pass

    @abstractmethod
    def write(self, filename: str) -> None:
        pass



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from typing import Tuple, Optional

import numpy as np

from miplearn.extractors.abstract import FeaturesExtractor
from miplearn.h5 import H5File


class AlvLouWeh2017Extractor(FeaturesExtractor):
    def __init__(
        self,
        with_m1: bool = True,
        with_m2: bool = True,
        with_m3: bool = True,
    ):
        self.with_m1 = with_m1
        self.with_m2 = with_m2
        self.with_m3 = with_m3

    def get_instance_features(self, h5: H5File) -> np.ndarray:
        raise NotImplemented()

    def get_var_features(self, h5: H5File) -> np.ndarray:
        """
        Computes static variable features described in:
            Alvarez, A. M., Louveaux, Q., & Wehenkel, L. (2017). A machine learning-based
            approximation of strong branching. INFORMS Journal on Computing, 29(1),
            185-195.
        """
        A = h5.get_sparse("static_constr_lhs")
        b = h5.get_array("static_constr_rhs")
        c = h5.get_array("static_var_obj_coeffs")
        c_sa_up = h5.get_array("lp_var_sa_obj_up")
        c_sa_down = h5.get_array("lp_var_sa_obj_down")
        values = h5.get_array("lp_var_values")

        assert A is not None
        assert b is not None
        assert c is not None

        nvars = len(c)
        curr = 0
        max_n_features = 40
        features = np.zeros((nvars, max_n_features))

        def push(v: np.ndarray) -> None:
            nonlocal curr
            assert v.shape == (nvars,), f"{v.shape} != ({nvars},)"
            features[:, curr] = v
            curr += 1

        def push_sign_abs(v: np.ndarray) -> None:
            assert v.shape == (nvars,), f"{v.shape} != ({nvars},)"
            push(np.sign(v))
            push(np.abs(v))

        def maxmin(M: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
            M_max = np.ravel(M.max(axis=0).todense())
            M_min = np.ravel(M.min(axis=0).todense())
            return M_max, M_min

        with np.errstate(divide="ignore", invalid="ignore"):
            # Feature 1
            push(np.sign(c))

            # Feature 2
            c_pos_sum = c[c > 0].sum()
            push(np.abs(c) / c_pos_sum)

            # Feature 3
            c_neg_sum = -c[c < 0].sum()
            push(np.abs(c) / c_neg_sum)

            if A is not None and self.with_m1:
                # Compute A_ji / |b_j|
                M1 = A.T.multiply(1.0 / np.abs(b)).T.tocsr()

                # Select rows with positive b_j and compute max/min
                M1_pos = M1[b > 0, :]
                if M1_pos.shape[0] > 0:
                    M1_pos_max = np.asarray(M1_pos.max(axis=0).todense()).flatten()
                    M1_pos_min = np.asarray(M1_pos.min(axis=0).todense()).flatten()
                else:
                    M1_pos_max = np.zeros(nvars)
                    M1_pos_min = np.zeros(nvars)

                # Select rows with negative b_j and compute max/min
                M1_neg = M1[b < 0, :]
                if M1_neg.shape[0] > 0:
                    M1_neg_max = np.asarray(M1_neg.max(axis=0).todense()).flatten()
                    M1_neg_min = np.asarray(M1_neg.min(axis=0).todense()).flatten()
                else:
                    M1_neg_max = np.zeros(nvars)
                    M1_neg_min = np.zeros(nvars)

                # Features 4-11
                push_sign_abs(M1_pos_min)
                push_sign_abs(M1_pos_max)
                push_sign_abs(M1_neg_min)
                push_sign_abs(M1_neg_max)

            if A is not None and self.with_m2:
                # Compute |c_i| / A_ij
                M2 = A.power(-1).multiply(np.abs(c)).tocsc()

                # Compute max/min
                M2_max, M2_min = maxmin(M2)

                # Make copies of M2 and erase elements based on sign(c)
                M2_pos_max = M2_max.copy()
                M2_neg_max = M2_max.copy()
                M2_pos_min = M2_min.copy()
                M2_neg_min = M2_min.copy()
                M2_pos_max[c <= 0] = 0
                M2_pos_min[c <= 0] = 0
                M2_neg_max[c >= 0] = 0
                M2_neg_min[c >= 0] = 0

                # Features 12-19
                push_sign_abs(M2_pos_min)
                push_sign_abs(M2_pos_max)
                push_sign_abs(M2_neg_min)
                push_sign_abs(M2_neg_max)

            if A is not None and self.with_m3:
                # Compute row sums
                S_pos = A.maximum(0).sum(axis=1)
                S_neg = np.abs(A.minimum(0).sum(axis=1))

                # Divide A by positive and negative row sums
                M3_pos = A.multiply(1 / S_pos).tocsr()
                M3_neg = A.multiply(1 / S_neg).tocsr()

                # Remove +inf and -inf generated by division by zero
                M3_pos.data[~np.isfinite(M3_pos.data)] = 0.0
                M3_neg.data[~np.isfinite(M3_neg.data)] = 0.0
                M3_pos.eliminate_zeros()
                M3_neg.eliminate_zeros()

                # Split each matrix into positive and negative parts
                M3_pos_pos = M3_pos.maximum(0)
                M3_pos_neg = -(M3_pos.minimum(0))
                M3_neg_pos = M3_neg.maximum(0)
                M3_neg_neg = -(M3_neg.minimum(0))

                # Calculate max/min
                M3_pos_pos_max, M3_pos_pos_min = maxmin(M3_pos_pos)
                M3_pos_neg_max, M3_pos_neg_min = maxmin(M3_pos_neg)
                M3_neg_pos_max, M3_neg_pos_min = maxmin(M3_neg_pos)
                M3_neg_neg_max, M3_neg_neg_min = maxmin(M3_neg_neg)

                # Features 20-35
                push_sign_abs(M3_pos_pos_max)
                push_sign_abs(M3_pos_pos_min)
                push_sign_abs(M3_pos_neg_max)
                push_sign_abs(M3_pos_neg_min)
                push_sign_abs(M3_neg_pos_max)
                push_sign_abs(M3_neg_pos_min)
                push_sign_abs(M3_neg_neg_max)
                push_sign_abs(M3_neg_neg_min)

            # Feature 36: only available during B&B

            # Feature 37
            if values is not None:
                push(
                    np.minimum(
                        values - np.floor(values),
                        np.ceil(values) - values,
                    )
                )

            # Features 38-43: only available during B&B

            # Feature 44
            if c_sa_up is not None:
                assert c_sa_down is not None

                # Features 44 and 46
                push(np.sign(c_sa_up))
                push(np.sign(c_sa_down))

                # Feature 45 is duplicated

                # Feature 47-48
                push(np.log(c - c_sa_down / np.sign(c)))
                push(np.log(c - c_sa_up / np.sign(c)))

                # Features 49-64: only available during B&B

        features = features[:, 0:curr]
        _fix_infinity(features)
        return features

    def get_constr_features(self, h5: H5File) -> np.ndarray:
        raise NotImplemented()


def _fix_infinity(m: Optional[np.ndarray]) -> None:
    if m is None:
        return
    masked = np.ma.masked_invalid(m)  # type: ignore
    max_values = np.max(masked, axis=0)
    min_values = np.min(masked, axis=0)
    m[:] = np.maximum(np.minimum(m, max_values), min_values)
    m[~np.isfinite(m)] = 0.0



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from math import pi
from typing import List, Optional

import gurobipy as gp
import numpy as np
from gurobipy import GRB, quicksum
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen

from miplearn.solvers.gurobi import GurobiModel


@dataclass
class UnitCommitmentData:
    demand: np.ndarray
    min_power: np.ndarray
    max_power: np.ndarray
    min_uptime: np.ndarray
    min_downtime: np.ndarray
    cost_startup: np.ndarray
    cost_prod: np.ndarray
    cost_fixed: np.ndarray


class UnitCommitmentGenerator:
    def __init__(
        self,
        n_units: rv_frozen = randint(low=1_000, high=1_001),
        n_periods: rv_frozen = randint(low=72, high=73),
        max_power: rv_frozen = uniform(loc=50, scale=450),
        min_power: rv_frozen = uniform(loc=0.5, scale=0.25),
        cost_startup: rv_frozen = uniform(loc=0, scale=10_000),
        cost_prod: rv_frozen = uniform(loc=0, scale=50),
        cost_fixed: rv_frozen = uniform(loc=0, scale=1_000),
        min_uptime: rv_frozen = randint(low=2, high=8),
        min_downtime: rv_frozen = randint(low=2, high=8),
        cost_jitter: rv_frozen = uniform(loc=0.75, scale=0.5),
        demand_jitter: rv_frozen = uniform(loc=0.9, scale=0.2),
        fix_units: bool = False,
    ) -> None:
        self.n_units = n_units
        self.n_periods = n_periods
        self.max_power = max_power
        self.min_power = min_power
        self.cost_startup = cost_startup
        self.cost_prod = cost_prod
        self.cost_fixed = cost_fixed
        self.min_uptime = min_uptime
        self.min_downtime = min_downtime
        self.cost_jitter = cost_jitter
        self.demand_jitter = demand_jitter
        self.fix_units = fix_units
        self.ref_data: Optional[UnitCommitmentData] = None

    def generate(self, n_samples: int) -> List[UnitCommitmentData]:
        def _sample() -> UnitCommitmentData:
            if self.ref_data is None:
                T = self.n_periods.rvs()
                G = self.n_units.rvs()

                # Generate unit parameteres
                max_power = self.max_power.rvs(G)
                min_power = max_power * self.min_power.rvs(G)
                max_power = max_power
                min_uptime = self.min_uptime.rvs(G)
                min_downtime = self.min_downtime.rvs(G)
                cost_startup = self.cost_startup.rvs(G)
                cost_prod = self.cost_prod.rvs(G)
                cost_fixed = self.cost_fixed.rvs(G)
                capacity = max_power.sum()

                # Generate periodic demand in the range [0.4, 0.8] * capacity, with a peak every 12 hours.
                demand = np.sin([i / 6 * pi for i in range(T)])
                demand *= uniform(loc=0, scale=1).rvs(T)
                demand -= demand.min()
                demand /= demand.max() / 0.4
                demand += 0.4
                demand *= capacity
            else:
                T, G = len(self.ref_data.demand), len(self.ref_data.max_power)
                demand = self.ref_data.demand * self.demand_jitter.rvs(T)
                min_power = self.ref_data.min_power
                max_power = self.ref_data.max_power
                min_uptime = self.ref_data.min_uptime
                min_downtime = self.ref_data.min_downtime
                cost_startup = self.ref_data.cost_startup * self.cost_jitter.rvs(G)
                cost_prod = self.ref_data.cost_prod * self.cost_jitter.rvs(G)
                cost_fixed = self.ref_data.cost_fixed * self.cost_jitter.rvs(G)

            data = UnitCommitmentData(
                demand.round(2),
                min_power.round(2),
                max_power.round(2),
                min_uptime,
                min_downtime,
                cost_startup.round(2),
                cost_prod.round(2),
                cost_fixed.round(2),
            )

            if self.ref_data is None and self.fix_units:
                self.ref_data = data

            return data

        return [_sample() for _ in range(n_samples)]


def build_uc_model(data: UnitCommitmentData) -> GurobiModel:
    """
    Models the unit commitment problem according to equations (1)-(5) of:

        Bendotti, P., Fouilhoux, P. & Rottner, C. The min-up/min-down unit
        commitment polytope. J Comb Optim 36, 1024-1058 (2018).
        https://doi.org/10.1007/s10878-018-0273-y

    """
    T = len(data.demand)
    G = len(data.min_power)
    D = data.demand
    Pmin, Pmax = data.min_power, data.max_power
    L = data.min_uptime
    l = data.min_downtime

    model = gp.Model()
    is_on = model.addVars(G, T, vtype=GRB.BINARY, name="is_on")
    switch_on = model.addVars(G, T, vtype=GRB.BINARY, name="switch_on")
    prod = model.addVars(G, T, name="prod")

    # Objective function
    model.setObjective(
        quicksum(
            is_on[g, t] * data.cost_fixed[g]
            + switch_on[g, t] * data.cost_startup[g]
            + prod[g, t] * data.cost_prod[g]
            for g in range(G)
            for t in range(T)
        )
    )

    # Eq 1: Minimum up-time constraint: If unit g is down at time t, then it
    # cannot have start up during the previous L[g] periods.
    model.addConstrs(
        (
            quicksum(switch_on[g, k] for k in range(t - L[g] + 1, t + 1)) <= is_on[g, t]
            for g in range(G)
            for t in range(L[g] - 1, T)
        ),
        name="eq_min_uptime",
    )

    # Eq 2: Minimum down-time constraint: Symmetric to the minimum-up constraint.
    model.addConstrs(
        (
            quicksum(switch_on[g, k] for k in range(t - l[g] + 1, t + 1))
            <= 1 - is_on[g, t - l[g] + 1]
            for g in range(G)
            for t in range(l[g] - 1, T)
        ),
        name="eq_min_downtime",
    )

    # Eq 3: Ensures that if unit g start up at time t, then the start-up variable
    # must be one.
    model.addConstrs(
        (
            switch_on[g, t] >= is_on[g, t] - is_on[g, t - 1]
            for g in range(G)
            for t in range(1, T)
        ),
        name="eq_startup",
    )

    # Eq 4: Ensures that demand is satisfied at each time period.
    model.addConstrs(
        (quicksum(prod[g, t] for g in range(G)) >= D[t] for t in range(T)),
        name="eq_demand",
    )

    # Eq 5: Sets the bounds to the quantity of power produced by each unit.
    model.addConstrs(
        (Pmin[g] * is_on[g, t] <= prod[g, t] for g in range(G) for t in range(T)),
        name="eq_prod_lb",
    )
    model.addConstrs(
        (prod[g, t] <= Pmax[g] * is_on[g, t] for g in range(G) for t in range(T)),
        name="eq_prod_ub",
    )
    model.update()

    return GurobiModel(model)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import List

import pytest

from miplearn.extractors.fields import H5FieldsExtractor
from miplearn.h5 import H5File


def test_fields_instance(multiknapsack_h5: List[str]) -> None:
    ext = H5FieldsExtractor(
        instance_fields=[
            "lp_obj_value",
            "lp_var_values",
            "static_var_obj_coeffs",
        ],
        var_fields=["lp_var_values"],
    )
    with H5File(multiknapsack_h5[0], "r") as h5:
        x = ext.get_instance_features(h5)
        assert x.shape == (201,)

        x = ext.get_var_features(h5)
        assert x.shape == (100, 1)


def test_fields_instance_none(multiknapsack_h5: List[str]) -> None:
    ext = H5FieldsExtractor(instance_fields=None)
    with H5File(multiknapsack_h5[0], "r") as h5:
        with pytest.raises(Exception):
            ext.get_instance_features(h5)



import logging
from miplearn.h5 import H5File
from glob import glob
import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score
import logging
from tqdm.auto import tqdm
import sys
import warnings

from sklearn.ensemble import GradientBoostingClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils import resample
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.dummy import DummyClassifier
from sklearn.model_selection import cross_val_score
from sklearn.multioutput import MultiOutputClassifier, ClassifierChain
from scipy import sparse
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.multiclass import unique_labels


logging.basicConfig(format="[%(asctime)s] %(message)s", level=logging.DEBUG)
logger = logging.getLogger()
# logger.setLevel(logging.DEBUG)
# stdout_handler = logging.StreamHandler(sys.stdout)
# stdout_handler.setLevel(logging.DEBUG)
# stdout_handler.setFormatter(logging.Formatter("[%(asctime)s] %(message)s"))
# logger.addHandler(stdout_handler)


class ConditionalClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, parent_clf, threshold=[0.01, 0.99]):
        self.parent_clf = parent_clf
        self.threshold = threshold
        self.mean_ = None

    def fit(self, x, y):
        if self.threshold[0] < y.mean() < self.threshold[1]:
            self.parent_clf.fit(x, y)
            self.classes_ = self.parent_clf.classes_
        else:
            self.mean_ = y.mean()
            self.classes_ = unique_labels(y)

    def predict(self, x):
        if self.mean_ is not None:
            return np.full(x.shape[0], np.round(self.mean_))
        else:
            return self.parent_clf.predict(x)


def run(max_samples=1_000_000):
    logger.info("Reading training data...")
    X, y = [], []
    for h5_filename in tqdm(
        glob("benchmark/data/setcover-n2750-m110/*.h5"), desc="read"
    ):
        with H5File(h5_filename, "r") as h5:
            var_values = h5.get_array("mip_var_values")
            var_features = h5.get_array("lp_var_features")
            var_types = h5.get_array("static_var_types")
            assert var_values is not None
            assert var_features is not None
            assert var_types is not None

            bin_var_indices = np.where(var_types == b"B")[0]
            X.append(var_features[bin_var_indices, :])
            y.append(var_values[bin_var_indices])

    logger.info("Constructing matrices...")
    X = np.vstack(X)
    y = np.hstack(y)
    assert len(X.shape) == 2
    n_samples, n_features = X.shape
    assert y.shape == (n_samples,)
    logger.info(f"Dataset has {n_samples:,d} samples and {n_features:,d} features")

    if n_samples > max_samples:
        logger.info(f"Too many samples. Resampling...")
        X, y = resample(X, y, n_samples=max_samples)
        n_samples, n_features = X.shape
        logger.info(f"Dataset has {n_samples:,d} samples and {n_features:,d} features")

    candidates = {
        "Dummy": DummyClassifier(),
        "Logistic": make_pipeline(
            StandardScaler(),
            LogisticRegression(),
        ),
        "GradientBoost": GradientBoostingClassifier(
            n_estimators=25,
            min_samples_split=10,
            max_depth=5,
        ),
    }

    best_clf = None
    best_clf_name = None
    best_score = -float("inf")

    for (clf_name, clf) in candidates.items():
        logger.info(f"Evaluating {clf_name}...")
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ConvergenceWarning)
            score = np.mean(cross_val_score(clf, X, y, cv=3))
        logger.info(f"    score = {score:.5f}")

        if score > best_score:
            best_score = score
            best_clf_name, best_clf = clf_name, clf

    logger.info(f"Selecting {best_clf_name}")


def run2():
    def _extract(h5_filename):
        with H5File(h5_filename, "r") as h5:
            var_types = h5.get_array("static_var_types")
            x = np.hstack(
                [
                    h5.get_array("static_var_obj_coeffs"),
                    h5.get_array("static_constr_rhs"),
                    h5.get_array("lp_var_values"),
                ]
            )
            y = h5.get_array("mip_var_values")
            bin_var_indices = np.where(var_types == b"B")[0]
            return x, y[bin_var_indices]

    logger.info("Reading training data...")
    x, y = [], []
    for h5_filename in tqdm(
        glob("benchmark/data/setcover-n2750-m110/*.h5"), desc="read"
    ):
        x_sample, y_sample = _extract(h5_filename)
        x.append(x_sample)
        y.append(y_sample)

    logger.info("Constructing matrices...")
    x = np.array(x)
    y = np.array(y)
    print(x.shape)
    print(y.shape)
    assert len(x.shape) == 2
    assert len(y.shape) == 2
    n_samples, n_features = x.shape
    logger.info(f"Dataset has {n_samples:,d} samples and {n_features:,d} features")

    # if n_samples > max_samples:
    #     logger.info(f"Too many samples. Resampling...")
    #     X, y = resample(X, y, n_samples=max_samples)
    #     n_samples, n_features = X.shape
    #     logger.info(f"Dataset has {n_samples:,d} samples and {n_features:,d} features")

    # clf = DecisionTreeClassifier(
    #     min_samples_split=10,
    #     max_depth=5,
    # )
    clf = ClassifierChain(
        ConditionalClassifier(
            make_pipeline(
                StandardScaler(),
                LogisticRegression(),
            ),
        ),
    )
    clf.fit(x, y)
    y_pred = clf.predict(x)
    print(len(y_pred))
    print(sparse.csr_matrix(y[0, :]))
    print()
    print(sparse.csr_matrix(y_pred[0, :]))


class VarObjCoeffExtractor:
    def extract(self, h5):
        var_obj_coeffs = h5.get_array("static_var_obj_coeffs")
        assert var_obj_coeffs is not None
        return var_obj_coeffs


class TopKSolutions:
    def __init__(self, k):
        self.k = k

    def construct(self, y_proba, solutions, var_names):
        assert len(y_proba.shape) == 1
        assert len(y_proba) == len(solutions)
        top_sol_indices = np.argpartition(y_proba, -self.k)[-self.k :]
        return [solutions[i] for i in top_sol_indices]


def run_3(
    extractor=VarObjCoeffExtractor(),
    constructor=TopKSolutions(3),
):
    h5_filenames = glob("benchmark/data/setcover-n2750-m110/*.h5")
    n_samples = len(h5_filenames)

    logger.info("Reading training data...")
    x, y, n_features = [], [], None
    solutions, bin_var_names = [], None
    solution_to_idx = {}
    for h5_filename in tqdm(h5_filenames, desc="read"):
        with H5File(h5_filename, "r") as h5:
            # Read optimal solution
            var_values = h5.get_array("mip_var_values")
            var_types = h5.get_array("static_var_types")
            assert var_values is not None
            assert var_types is not None

            # Extract binary part
            bin_var_indices = np.where(var_types == b"B")[0]
            bin_var_values = var_values[bin_var_indices]

            # Extract variable names
            if bin_var_names is None:
                var_names = h5.get_array("static_var_names")
                assert var_names is not None
                bin_var_names = var_names[bin_var_indices]

            # Store solution
            sol = tuple(np.where(bin_var_values)[0])
            if sol not in solution_to_idx:
                solutions.append(
                    {
                        var_name: bin_var_values[var_idx]
                        for (var_idx, var_name) in enumerate(bin_var_names)
                    }
                )
                solution_to_idx[sol] = len(solution_to_idx)
            y.append(solution_to_idx[sol])

            # Extract features
            x_sample = extractor.extract(h5)
            assert len(x_sample.shape) == 1
            if n_features is None:
                n_features = len(x_sample)
            else:
                assert len(x_sample) == n_features
            x.append(x_sample)

    logger.info("Constructing matrices...")
    x = np.vstack(x)
    y = np.array(y)
    assert len(x.shape) == 2
    assert x.shape[0] == n_samples
    assert x.shape[1] == n_features
    assert y.shape == (n_samples,)
    n_classes = len(solution_to_idx)
    logger.info(
        f"Dataset has {n_samples:,d} samples, {n_features:,d} features and {n_classes:,d} classes"
    )

    clf = DecisionTreeClassifier(max_depth=3)
    clf.fit(x, y)
    y_proba = clf.predict_proba(x)

    sols = constructor.construct(y_proba[0, :], solutions, var_names)


run_3()



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from os.path import exists

from config import challenges, n_jobs
from miplearn.io import save

if __name__ == "__main__":
    for challenge in challenges:
        if exists(f"data/{challenge.name}/"):
            print(f"Skipping data/{challenge.name}")
            continue
        print(f"\n{challenge.name}")
        data = challenge.generator.generate(challenge.n_samples)
        save(
            data,
            f"data/{challenge.name}/",
            prefix=f"{challenge.name}-",
            n_jobs=n_jobs,
        )



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
import numpy as np

from miplearn.extractors.abstract import FeaturesExtractor
from miplearn.h5 import H5File


class DummyExtractor(FeaturesExtractor):
    def get_instance_features(self, h5: H5File) -> np.ndarray:
        return np.zeros(1)

    def get_var_features(self, h5: H5File) -> np.ndarray:
        var_types = h5.get_array("static_var_types")
        assert var_types is not None
        n_vars = len(var_types)
        return np.zeros((n_vars, 1))

    def get_constr_features(self, h5: H5File) -> np.ndarray:
        constr_sense = h5.get_array("static_constr_sense")
        assert constr_sense is not None
        n_constr = len(constr_sense)
        return np.zeros((n_constr, 1))



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import networkx as nx
import numpy as np

from miplearn.problems.vertexcover import (
    MinWeightVertexCoverData,
    build_vertexcover_model,
)


def test_stab() -> None:
    data = MinWeightVertexCoverData(
        graph=nx.cycle_graph(5),
        weights=np.array([1.0, 1.0, 1.0, 1.0, 1.0]),
    )
    model = build_vertexcover_model(data)
    model.optimize()
    assert model.inner.objVal == 3.0



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from io import StringIO
from typing import Callable

import gurobipy as gp
from gurobipy import GRB, LinExpr

from ..h5 import H5File
from ..io import _RedirectOutput, load

import numpy as np


class LazyCollector:
    def __init__(
        self,
        min_constrs: int = 100_000,
        time_limit: float = 900,
    ) -> None:
        self.min_constrs = min_constrs
        self.time_limit = time_limit

    def collect(self, data: str, build_model: Callable, tol: float = 1e-6) -> None:
        h5_filename = data.replace(".pkl.gz", ".h5")
        with H5File(h5_filename, "r+") as h5:
            streams = [StringIO()]
            lazy = None
            with _RedirectOutput(streams):
                slacks = h5.get_array("mip_constr_slacks")
                assert slacks is not None

                # Check minimum problem size
                if len(slacks) < self.min_constrs:
                    print("Problem is too small. Skipping.")
                    h5.put_array("mip_constr_lazy", np.zeros(len(slacks)))
                    return

                # Load model
                print("Loading model...")
                model = load(data, build_model)
                model.params.LazyConstraints = True
                model.params.timeLimit = self.time_limit
                gp_constrs = np.array(model.getConstrs())
                gp_vars = np.array(model.getVars())

                # Load constraints
                lhs = h5.get_sparse("static_constr_lhs")
                rhs = h5.get_array("static_constr_rhs")
                sense = h5.get_array("static_constr_sense")
                assert lhs is not None
                assert rhs is not None
                assert sense is not None
                lhs_csr = lhs.tocsr()
                lhs_csc = lhs.tocsc()
                constr_idx = np.array(range(len(rhs)))
                lazy = np.zeros(len(rhs))

                # Drop loose constraints
                selected = (slacks > 0) & ((sense == b"<") | (sense == b">"))
                loose_constrs = gp_constrs[selected]
                print(
                    f"Removing {len(loose_constrs):,d} constraints (out of {len(rhs):,d})..."
                )
                model.remove(list(loose_constrs))

                # Filter to constraints that were dropped
                lhs_csr = lhs_csr[selected, :]
                lhs_csc = lhs_csc[selected, :]
                rhs = rhs[selected]
                sense = sense[selected]
                constr_idx = constr_idx[selected]
                lazy[selected] = 1

                # Load warm start
                var_names = h5.get_array("static_var_names")
                var_values = h5.get_array("mip_var_values")
                assert var_values is not None
                assert var_names is not None
                for (var_idx, var_name) in enumerate(var_names):
                    var = model.getVarByName(var_name.decode())
                    var.start = var_values[var_idx]

                print("Solving MIP with lazy constraints callback...")

                def callback(model: gp.Model, where: int) -> None:
                    assert rhs is not None
                    assert lazy is not None
                    assert sense is not None

                    if where == GRB.Callback.MIPSOL:
                        x_val = np.array(model.cbGetSolution(model.getVars()))
                        slack = lhs_csc * x_val - rhs
                        slack[sense == b">"] *= -1
                        is_violated = slack > tol

                        for (j, rhs_j) in enumerate(rhs):
                            if is_violated[j]:
                                lazy[constr_idx[j]] = 0
                                expr = LinExpr(
                                    lhs_csr[j, :].data, gp_vars[lhs_csr[j, :].indices]
                                )
                                if sense[j] == b"<":
                                    model.cbLazy(expr <= rhs_j)
                                elif sense[j] == b">":
                                    model.cbLazy(expr >= rhs_j)
                                else:
                                    raise RuntimeError(f"Unknown sense: {sense[j]}")

                model.optimize(callback)
                print(f"Marking {lazy.sum():,.0f} constraints as lazy...")

            h5.put_array("mip_constr_lazy", lazy)
            h5.put_scalar("mip_constr_lazy_log", streams[0].getvalue())
